package edu.sc.seis.cormorant.plottable;

import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.omg.CORBA.NO_IMPLEMENT;
import org.omg.CORBA.NO_MEMORY;
import org.omg.CORBA.UNKNOWN;

import edu.iris.Fissures.Dimension;
import edu.iris.Fissures.Plottable;
import edu.iris.Fissures.Time;
import edu.iris.Fissures.IfEvent.EventAccess;
import edu.iris.Fissures.IfNetwork.ChannelId;
import edu.iris.Fissures.IfPlottable.PlottableDCPOA;
import edu.iris.Fissures.IfPlottable.UnsupportedDimension;
import edu.iris.Fissures.IfSeismogramDC.DataCenterOperations;
import edu.iris.Fissures.IfSeismogramDC.LocalSeismogram;
import edu.iris.Fissures.IfSeismogramDC.RequestFilter;
import edu.iris.Fissures.model.MicroSecondDate;
import edu.iris.Fissures.model.TimeInterval;
import edu.iris.Fissures.model.UnitImpl;
import edu.iris.Fissures.network.ChannelIdUtil;
import edu.iris.Fissures.network.NetworkAttrImpl;
import edu.iris.Fissures.seismogramDC.LocalSeismogramImpl;
import edu.iris.Fissures.seismogramDC.RequestFilterUtil;
import edu.sc.seis.fissuresUtil.cache.BulletproofVestFactory;
import edu.sc.seis.fissuresUtil.chooser.ClockUtil;
import edu.sc.seis.fissuresUtil.database.ConnMgr;
import edu.sc.seis.fissuresUtil.database.ConnectionCreator;
import edu.sc.seis.fissuresUtil.database.NotFound;
import edu.sc.seis.fissuresUtil.display.MicroSecondTimeRange;
import edu.sc.seis.fissuresUtil.display.SimplePlotUtil;
import edu.sc.seis.fissuresUtil.exceptionHandler.GlobalExceptionHandler;
import edu.sc.seis.fissuresUtil.hibernate.NetworkDB;
import edu.sc.seis.fissuresUtil.hibernate.PlottableChunk;
import edu.sc.seis.fissuresUtil.hibernate.PlottableDB;
import edu.sc.seis.fissuresUtil.simple.Initializer;
import edu.sc.seis.fissuresUtil.time.CoverageTool;

public class PlottableImpl extends PlottableDCPOA {

    public PlottableImpl(ConnectionCreator connCreator, String dcDNS, String dcName, int dcTries) {
        this(getDataCenter(dcDNS, dcName, dcTries), connCreator);
    }

    public PlottableImpl(boolean cacheOnly, String dcDNS, String dcName, int dcTries, ConnectionCreator connCreator) {
        this((cacheOnly ? null : getDataCenter(dcDNS, dcName, dcTries)), connCreator);
    }

    public PlottableImpl(DataCenterOperations dc, ConnectionCreator connCreator) {
        this.dataCenter = dc;
        plottableDB = PlottableDB.getSingleton();
        netDB = NetworkDB.getSingleton();
    }

    public boolean custom_sizes() {
        return false;
    }

    public Plottable[] get_plottable(RequestFilter request, Dimension pixel_size)
            throws UnsupportedDimension {
        // due to updating db, don't want to work on same channel in more than one thread
        String chanWorkingOn = ChannelIdUtil.toStringNoDates(request.channel_id).intern();
        synchronized(chanWorkingOn) {
        long requestNumber = getRequestCount();
        logger.info(requestNumber + ": got request "
                + RequestFilterUtil.toString(request) + " at pixel size "
                + pixel_size.width);
        MicroSecondDate start = ClockUtil.now();
        int width = pixel_size.width;
        MicroSecondTimeRange requestRange = new MicroSecondTimeRange(request);
        int requestedPixelsPerDay = (int)(Math.floor(getDays(requestRange)) * width);
        boolean acceptableWidth = false;
        for(int i = 0; i < DAY_DIMS.length; i++) {
            if(DAY_DIMS[i].width == requestedPixelsPerDay) {
                acceptableWidth = true;
            }
        }
        if(!acceptableWidth) {
            throw new UnsupportedDimension(getCacheOnlyForLogger()
                    + " request #" + requestNumber
                    + ": Acceptable pixels per day are " + SUPPORTED_WIDTHS);
        }
        try {
            // start looking for better chanId than the one in the request
            // (IE we don't care about the start dates)
            // this may break if we hook this up to a seismogram server that
            // actually cares about dates.
            ChannelId chanId = getBestChannelIdFromDb(request);
            request.channel_id = chanId;
            RequestFilter[] filters = {request};
            // end of that long mess
            List<PlottableChunk> chunks;
                chunks = plottableDB.get(requestRange,
                                         chanId.network_id.network_code,
                                         chanId.station_code,
                                         chanId.site_code,
                                         chanId.channel_code,
                                         (int)(Math.floor(getDays(requestRange)) * width));
            
            logger.debug(requestNumber + ": Got " + chunks.size()
                    + " chunks for request #" + requestNumber + " "
                    + requestRange);
            MicroSecondTimeRange[] ranges = new MicroSecondTimeRange[chunks.size()];
            TimeInterval secondsNotCovered = new TimeInterval(0, UnitImpl.SECOND);
            for(int i = 0; i < ranges.length; i++) {
                ranges[i] = chunks.get(i).getTimeRange();
            }
            filters = CoverageTool.notCoveredIgnoreGaps(filters, ranges);
            for (int i = 0; i < filters.length; i++) {
                MicroSecondDate rfStart = new MicroSecondDate(filters[i].start_time);
                MicroSecondDate rfEnd = new MicroSecondDate(filters[i].end_time);
                secondsNotCovered = secondsNotCovered.add(rfEnd.subtract(rfStart));
            }
            List<PlottableChunk> newPlots = new ArrayList<PlottableChunk>();
            if(filters.length > 0 && dataCenter != null && secondsNotCovered.greaterThan(SMALL_REQUEST_TOL)) {
                for(int i = 0; i < filters.length; i++) {
                    logger.debug(requestNumber + ": "
                            + RequestFilterUtil.toString(filters[i])
                            + " not covered by data in db");
                }
                // because some seismogram servers can't handle large requests,
                // we're breaking them up into requests <= six hours
                filters = breakUpFilters(filters,
                                         new TimeInterval(6.0, UnitImpl.HOUR));
                for(int i = 0; i < filters.length; i++) {
                    try {
                        logger.debug(requestNumber + ": request filter # " + i
                                + " to data center: "
                                + RequestFilterUtil.toString(filters[i]));
                        LocalSeismogram[] seis = dataCenter.retrieve_seismograms(new RequestFilter[] {filters[i]});
                        for(int j = 0; j < seis.length; j++) {
                            LocalSeismogramImpl impl = (LocalSeismogramImpl)seis[j];
                            logger.debug(requestNumber
                                    + ": Got seismogram from "
                                    + new MicroSecondTimeRange(impl));
                        }
                        for(int j = 0; j < DAY_DIMS.length; j++) {
                            int curWidth = DAY_DIMS[j].width;
                            List<PlottableChunk> seisChunks = SimplePlotUtil.makePlottables(seis,
                                                                         curWidth);
                            PlottableDB.getSingleton().put(seisChunks);
                            for (PlottableChunk chunk : seisChunks) {
                                if (chunk.getPixelsPerDay() == requestedPixelsPerDay) {
                                    newPlots.add(chunk);
                                }
                            }
                        }
                    } catch(NO_MEMORY e) {
                        logger.error(requestNumber + ": request filter # " + i
                                + " to data center: "
                                + RequestFilterUtil.toString(filters[i]), e);
                        throw e;
                    }
                }
                PlottableDB.commit();
            }

            
            List<PlottableChunk> all = new ArrayList<PlottableChunk>(chunks.size() + newPlots.size());
            all.addAll(chunks);
            all.addAll(newPlots);
            // need to convert all plottables to common pixel scale, ie all
            // start at 0 being their beginning
            all = convertToCommonPixelScale(all, requestRange, requestedPixelsPerDay);
            
            MicroSecondDate stop = ClockUtil.now();
            logger.debug(requestNumber + ": "
                    + stop.difference(start).convertTo(UnitImpl.SECOND)
                    + " to satisfy request for " + requestRange);
            Plottable[] out = new Plottable[all.size()];
            for (int i = 0; i < out.length; i++) {
                out[i] = all.get(i).getData();
            }
            // rollback so read only connections can be returned to pool, above commit handles cases where db is changed
            PlottableDB.rollback();
            logger.debug("finish request: "+requestNumber);
            return out;
        } catch(Throwable e) {
            GlobalExceptionHandler.handle("Exception occured while retrieving plottable using get_plottable. request "
                                                  + requestNumber
                                                  + ": "
                                                  + ChannelIdUtil.toString(request.channel_id)
                                                  + " from "
                                                  + new MicroSecondTimeRange(request.start_time,
                                                                             request.end_time),
                                          e);
            PlottableDB.rollback();
            throw new UNKNOWN(e.toString());
        }
        }
    }

    protected List<PlottableChunk> convertToCommonPixelScale(List<PlottableChunk> chunks,
                                                             MicroSecondTimeRange requestRange,
                                                             int pixelsPerDay) {
        int requestPixels = getPixels(pixelsPerDay, requestRange);
        List<PlottableChunk> outChunks = new ArrayList<PlottableChunk>();
        Iterator<PlottableChunk> it = chunks.iterator();
        while (it.hasNext()) {
            PlottableChunk pc = it.next();
            MicroSecondDate rowBeginTime = pc.getBeginTime();
            int offsetIntoRequestPixels = SimplePlotUtil.getPixel(requestPixels, requestRange, rowBeginTime);
            int numPixels = pc.getNumPixels();
            int firstPixelForRequest = 0;
            if (offsetIntoRequestPixels < 0) {
                // This db row has data starting before the request, start
                // at
                // pertinent point
                firstPixelForRequest = -1 * offsetIntoRequestPixels;
            }
            int lastPixelForRequest = numPixels;
            if (offsetIntoRequestPixels + numPixels > requestPixels) {
                // This row has more data than was requested in it, only get
                // enough to fill the request
                lastPixelForRequest = requestPixels - offsetIntoRequestPixels;
            }
            if (firstPixelForRequest > lastPixelForRequest) {
                throw new NegativeArraySizeException("first pixel > last pixel: f="+firstPixelForRequest+"  l="+lastPixelForRequest);
            }
            int pixelsUsed = lastPixelForRequest - firstPixelForRequest;
            int[] x = new int[pixelsUsed * 2];
            int[] y = new int[pixelsUsed * 2];
            int[] ploty = pc.getYData();
            System.arraycopy(ploty, firstPixelForRequest*2, y, 0, pixelsUsed*2);
            for (int i = 0; i < pixelsUsed * 2; i++) {
                x[i] = firstPixelForRequest + offsetIntoRequestPixels + i / 2;
            }
            Plottable p = new Plottable(x, y);
            PlottableChunk shiftPC = new PlottableChunk(p,
                                                        PlottableChunk.getPixel(rowBeginTime, pixelsPerDay)
                                                                + firstPixelForRequest,
                                                        PlottableChunk.getJDay(rowBeginTime),
                                                        PlottableChunk.getYear(rowBeginTime),
                                                        pixelsPerDay,
                                                        pc.getNetworkCode(),
                                                        pc.getStationCode(),
                                                        pc.getSiteCode(),
                                                        pc.getChannelCode());
            outChunks.add(shiftPC);
        }
        return outChunks;
    }

    /**
     * look for better chanId than the one in the request (IE we don't care
     * about the start dates). this may break if we hook this up to a seismogram
     * server that actually cares about dates.
     */
    public ChannelId getBestChannelIdFromDb(RequestFilter request) {
        try {
            NetworkAttrImpl attr = netDB.getNetworkById(request.channel_id.network_id);
            if (attr != null) {
                return netDB.getChannel(attr.get_code(),
                                        request.channel_id.station_code,
                                        request.channel_id.site_code,
                                        request.channel_id.channel_code,
                                        new MicroSecondDate(request.start_time)).get_id();
            }
        } catch(NotFound e) {
            // NotFound, so not already in db, use the id in the request
        }
        return request.channel_id;
    }

    protected static List<PlottableChunk> commitPlotsToDBAndFilterForWidth(List<PlottableChunk> chunks,
                                                                  int pixelsPerDay,
                                                                  PlottableDB db) throws IOException {
        
        boolean foundDayDim = false;
        for (int i = 0; i < DAY_DIMS.length; i++) {
            if (DAY_DIMS[i].width == pixelsPerDay) {
                foundDayDim = true;
            }
        }
        if (! foundDayDim) {
            return new ArrayList<PlottableChunk>();
        }
        List<PlottableChunk> plottable = new ArrayList<PlottableChunk>();
        for (PlottableChunk pc : chunks) {
            if (pc.getData().x_coor.length > 0) {
                plottable.add(pc);
            }
        }
            db.put(plottable);
            db.commit();
        return plottable;
    }
    
    public Plottable[] get_for_day(ChannelId channel_id, int year, int jday, Dimension pixel_size)
            throws UnsupportedDimension {
        try {
            Time[] times = getTimes(jday, year);
            return get_plottable(new RequestFilter(channel_id, times[0], times[1]), pixel_size);
        } catch(UnsupportedDimension e) {
            throw e;
        } catch(Throwable e) {
            GlobalExceptionHandler.handle(getCacheOnlyForLogger()
                    + "Exception occured while retrieving plottable using get_for_day", e);
            throw new org.omg.CORBA.UNKNOWN(e.toString());
        }
    }

    public Dimension[] get_whole_day_sizes() {
        return DAY_DIMS;
    }

    private static double getDays(MicroSecondTimeRange tr) {
        return tr.getInterval().convertTo(UnitImpl.DAY).getValue();
    }

    private static final Time[] getTimes(int jday, int year) {
        DecimalFormat numberFormat = new DecimalFormat("000");
        String yearString = new Integer(year).toString();
        Time startTime = new Time(yearString + numberFormat.format(jday) + "J00:00:00.000Z", -1);
        Time endTime = new Time(yearString + numberFormat.format(jday + 1) + "J00:00:00.000Z", -1);
        return new Time[] {startTime, endTime};
    }

    public Dimension[] get_event_sizes() {
        throw new NO_IMPLEMENT();
    }

    public Plottable[] get_for_event(EventAccess event, ChannelId channel_id, Dimension pixel_size) {
        throw new NO_IMPLEMENT();
    }

    private static DataCenterOperations getDataCenter(String dns, String name, int tries) {
        return BulletproofVestFactory.vestSeismogramDC(dns, name, Initializer.getNS(), tries);
    }

    private String getCacheOnlyForLogger() {
        if (dataCenter == null) {
            return "CachePlottableImpl: ";
        } else {
            return "DCPlottableImpl: ";
        }
    }

    private long getRequestCount() {
        if (requestCount <= Long.MAX_VALUE) {
            return requestCount++;
        } else {
            requestCount = 0;
        }
        return 0;
    }

    public static RequestFilter[] breakUpFilters(RequestFilter[] filters, TimeInterval maxRequestLength) {
        List okayFilters = new ArrayList();
        for (int i = 0; i < filters.length; i++) {
            RequestFilter curFilter = filters[i];
            MicroSecondDate start = new MicroSecondDate(curFilter.start_time);
            MicroSecondDate end = new MicroSecondDate(curFilter.end_time);
            TimeInterval reqLen;
            do {
                reqLen = end.subtract(start);
                MicroSecondDate breakPoint = end;
                if (reqLen.greaterThan(maxRequestLength)) {
                    breakPoint = start.add(maxRequestLength);
                }
                RequestFilter req = new RequestFilter(curFilter.channel_id,
                                                      start.getFissuresTime(),
                                                      breakPoint.getFissuresTime());
                okayFilters.add(req);
                curFilter.start_time = breakPoint.getFissuresTime();
                start = breakPoint;
            } while (reqLen.greaterThan(maxRequestLength));
        }
        return (RequestFilter[])okayFilters.toArray(new RequestFilter[0]);
    }

    private static int getPixels(int pixelsPerDay, MicroSecondTimeRange tr) {
        TimeInterval inter = tr.getInterval();
        inter = (TimeInterval)inter.convertTo(UnitImpl.DAY);
        double samples = pixelsPerDay * inter.getValue();
        return (int)Math.floor(samples);
    }

    private DataCenterOperations dataCenter;

    private static final Dimension[] DAY_DIMS = new Dimension[] {new Dimension(5400, 1), new Dimension(6000, 1)};

    public static Dimension[] getDayDims() {
        return DAY_DIMS;
    }
    
    private static final String SUPPORTED_WIDTHS;
    static {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < DAY_DIMS.length; i++) {
            sb.append(DAY_DIMS[i].width + " ");
        }
        SUPPORTED_WIDTHS = sb.toString();
    }

    private static final Logger logger = LoggerFactory.getLogger(PlottableImpl.class);

    private PlottableDB plottableDB;

    private NetworkDB netDB;

    private TimeInterval SMALL_REQUEST_TOL = new TimeInterval(15, UnitImpl.MINUTE);
    
    private long requestCount = 0;
}
